/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */

/************************************************************************
 *
 * File: %M%
 *
 * Copyright (c) 1994 by Silicon Graphics Inc., Mt. View, CA 94039
 *
 * Created:       Mon Jul 11 16:39:20 1994 by George Kyriazis
 * Last Modified: Tue Jul 12 13:40:18 1994 by George Kyriazis
 *
 * how to render a cone
 *
 ************************************************************************/

#ident "$Revision: 1.1 $"

#define private protected
#include "GlCone.h"
#undef private

#include <Inventor/actions/SoGLRenderAction.h>


// Shorthand for testing whether current parts includes PART
#define HAS_PART(PARTS, PART)	((PARTS & (PART)) != 0)

// Returns S or T texture coord for point on bottom of cone, given x
// or z coord
#define BOT_TEX_S(x)	((x) * .5 + .5)
#define BOT_TEX_T(z)	((z) * .5 + .5)

SO_NODE_SOURCE(GlCone);

GlCone::GlCone()
{
    SO_NODE_CONSTRUCTOR(GlNode);
    isBuiltIn = TRUE;
}

GlCone::~GlCone()
{
}


void
GlCone::initClass()
{
    classTypeId = SoType::overrideType(SoCone::getClassTypeId(),
				       &GlCone::createInstance);
    parentFieldData = SoCone::getFieldDataPtr();
}


void
GlCone::GLRender(SoGLRenderAction *action)
{
    SbBool	unitSize;
    float	radius, halfHeight;

    // First see if the object is visible and should be rendered now
    if (! shouldGLRender(action))
	return;

    // Scale world if necessary
    getSize(radius, halfHeight);
    unitSize = (radius == 1.0 && halfHeight == 1.0);
    if (! unitSize) {
	glPushMatrix();
	glScalef(radius, halfHeight, radius);
    }

    // Render the cone.
    GLRenderGeneric(action);

    if (! unitSize)
	glPopMatrix();
}

void
GlCone::GLRenderGeneric(SoGLRenderAction *action)
//
////////////////////////////////////////////////////////////////////////
{
    int			curParts, numSides, numSections, side, section;
    float		yTop, yBot, dy;
    float		s, ds, tTop, tBot, dt;
    float		outerRadius, innerRadius, dRadius;
    SbVec2f		*baseCoords;
    SbVec3f		*sideNormals, pt, norm;

    curParts = (parts.isIgnored() ? ALL : parts.getValue());

    // Compute number of sides and sections to use to represent
    // cone, then compute ring of x,z coordinates around cone
    // and store in baseCoords.
    computeBase(action, numSides, numSections, baseCoords, sideNormals);

    dRadius = 1.0 / numSections;

    if (HAS_PART(curParts, SIDES)) {

	// Draw each section of sides as a triangle mesh, from top to bottom
	yTop = 1.0;
	dy   = -2.0 / numSections;
	tTop = 1.0;
	dt   = -1.0 / numSections;
	ds   =  1.0 / numSides;

	innerRadius = 0.0;

	for (section = 0; section < numSections; section++) {

	    outerRadius = innerRadius + dRadius;
	    yBot = yTop + dy;

	    tBot = tTop + dt;
	    s    = 0.0;

	    glBegin(GL_TRIANGLE_STRIP);

	    for (side = 0; side < numSides; side++) {
		// Deal with normal
		glNormal3fv(sideNormals[side].getValue());

		// Point at bottom of section
		pt.setValue(outerRadius * baseCoords[side][0], yBot,
			    outerRadius * baseCoords[side][1]);
		glTexCoord2f(s, tBot);
		glVertex3fv(pt.getValue());

		// Point at top of section
		pt.setValue(innerRadius * baseCoords[side][0], yTop,
			    innerRadius * baseCoords[side][1]);
		glTexCoord2f(s, tTop);
		glVertex3fv(pt.getValue());
		s += ds;
	    }

	    // Join end of strip back to beginning
	    side = 0;
	    s = 1.0;
	    // Deal with normal
	    glNormal3fv(sideNormals[side].getValue());

	    // Point at bottom of section
	    pt.setValue(outerRadius * baseCoords[side][0], yBot,
			outerRadius * baseCoords[side][1]);
	    glTexCoord2f(s, tBot);
	    glVertex3fv(pt.getValue());

	    // Point at top of section
	    pt.setValue(innerRadius * baseCoords[side][0], yTop,
			innerRadius * baseCoords[side][1]);
	    glTexCoord2f(s, tTop);
	    glVertex3fv(pt.getValue());
	    s += ds;

	    glEnd();

	    // Prepare for next section down
	    innerRadius = outerRadius;
	    yTop = yBot;
	    tTop = tBot;
	}
    }

    // Draw bottom face as a series of concentric rings. The number of
    // rings is the same as the number of sections of the sides of the
    // cone.
    if (HAS_PART(curParts, BOTTOM)) {
	norm.setValue(0.0, -1.0, 0.0);
	pt[1] = -1.0;

	glNormal3fv(norm.getValue());

	// Start at the outside and work in
	outerRadius = 1.0;
	for (section = numSections - 1; section >= 0; --section) {

	    innerRadius = outerRadius - dRadius;

	    // Innermost ring is drawn as a triangle fan. This not
	    // only gets better shading (because the center vertex is
	    // sent), but also avoids the problem of having a polygon
	    // with too many vertices.
	    if (section == 0) {
		glBegin(GL_TRIANGLE_FAN);

		// Center point comes first
		pt[0] = pt[2] = 0.0;
		glTexCoord2f(0.5, 0.5);
		glVertex3fv(pt.getValue());

		// Send all vertices around ring
		for (side = 0; side < numSides; side++) {
		    pt[0] = outerRadius * baseCoords[side][0];
		    pt[2] = outerRadius * baseCoords[side][1];
		    glTexCoord2f(BOT_TEX_S(pt[0]), BOT_TEX_T(pt[2]));
		    glVertex3fv(pt.getValue());
		}
		// Send first vertex again
		pt[0] = outerRadius * baseCoords[0][0];
		pt[2] = outerRadius * baseCoords[0][1];
		glTexCoord2f(BOT_TEX_S(pt[0]), BOT_TEX_T(pt[2]));
		glVertex3fv(pt.getValue());

		glEnd();
	    }

	    // Other rings are triangle strips
	    else {
		glBegin(GL_TRIANGLE_STRIP);

		// Go in reverse order so that vertex ordering is correct
		for (side = numSides - 1; side >= 0; side--) {
		    // Send points on outer and inner rings
		    pt[0] = outerRadius * baseCoords[side][0];
		    pt[2] = outerRadius * baseCoords[side][1];
		    glTexCoord2f(BOT_TEX_S(pt[0]), BOT_TEX_T(pt[2]));
		    glVertex3fv(pt.getValue());
		    pt[0] = innerRadius * baseCoords[side][0];
		    pt[2] = innerRadius * baseCoords[side][1];
		    glTexCoord2f(BOT_TEX_S(pt[0]), BOT_TEX_T(pt[2]));
		    glVertex3fv(pt.getValue());
		}

		// Join end of strip back to beginning
		side = numSides - 1;
		pt[0] = outerRadius * baseCoords[side][0];
		pt[2] = outerRadius * baseCoords[side][1];
		glTexCoord2f(BOT_TEX_S(pt[0]), BOT_TEX_T(pt[2]));
		glVertex3fv(pt.getValue());
		pt[0] = innerRadius * baseCoords[side][0];
		pt[2] = innerRadius * baseCoords[side][1];
		glTexCoord2f(BOT_TEX_S(pt[0]), BOT_TEX_T(pt[2]));
		glVertex3fv(pt.getValue());

		glEnd();

		// Prepare for next ring
		outerRadius = innerRadius;
	    }
	}
    }
}

