/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */

/************************************************************************
 *
 * File: %M%
 *
 * Copyright (c) 1994 by Silicon Graphics Inc., Mt. View, CA 94039
 *
 * Created:       Tue Jul 12 12:47:19 1994 by George Kyriazis
 * Last Modified: Tue Jul 12 13:40:47 1994 by George Kyriazis
 *
 * how to render a cylinder
 *
 ************************************************************************/

#ident "$Revision: 1.1 $"

#define private protected
#include "GlCylinder.h"
#undef private

#include <Inventor/actions/SoGLRenderAction.h>


SO_NODE_SOURCE(GlCylinder);


// Shorthand for testing whether current parts includes PART
#define HAS_PART(PARTS, PART)	((PARTS & (PART)) != 0)

// Returns S or T texture coord for point on top or bottom of
// cylinder, given x or z coord
#define BOT_TEX_S(x)	((x) * .5 + .5)
#define BOT_TEX_T(z)	((z) * .5 + .5)
#define TOP_TEX_S(x)	BOT_TEX_S(x)
#define TOP_TEX_T(z)	(1.0 - BOT_TEX_T(z))


GlCylinder::GlCylinder()
{
    SO_NODE_CONSTRUCTOR(GlCylinder);
    isBuiltIn = TRUE;
}

GlCylinder::~GlCylinder()
{
}


void
GlCylinder::initClass()
{
    classTypeId = SoType::overrideType(SoCylinder::getClassTypeId(),
				       &GlCylinder::createInstance);
    parentFieldData = SoCylinder::getFieldDataPtr();
}


void
GlCylinder::GLRender(SoGLRenderAction *action)
{
    SbBool	unitSize;
    float	radius, halfHeight;

    // First see if the object is visible and should be rendered now
    if (! shouldGLRender(action))
	return;

    // Scale world if necessary
    getSize(radius, halfHeight);
    unitSize = (radius == 1.0 && halfHeight == 1.0);
    if (! unitSize) {
	glPushMatrix();
	glScalef(radius, halfHeight, radius);
    }

    // Render the cylinder. 
    GLRenderGeneric(action);

    if (! unitSize)
	glPopMatrix();
}

void
GlCylinder::GLRenderGeneric(SoGLRenderAction *action)
{
    int			curParts, numSides, numSections, side, section;
    float		yTop, yBot, dy;
    float		s, ds, tTop, tBot, dt;
    float		outerRadius, innerRadius, dRadius;
    SbVec2f		*ringCoords;
    SbVec3f		pt, norm;

    curParts = (parts.isIgnored() ? ALL : parts.getValue());

    // Compute number of sides and sections to use to represent
    // cylinder, then compute ring of x,z coordinates around cylinder
    // and store in ringCoords.
    computeRing(action, numSides, numSections, ringCoords);

    if (HAS_PART(curParts, SIDES)) {

	// Draw each section of sides as a triangle mesh, from top to bottom
	yTop = 1.0;
	dy   = -2.0 / numSections;
	tTop = 1.0;
	dt   = -1.0 / numSections;
	ds   =  1.0 / numSides;

	for (section = 0; section < numSections; section++) {

	    yBot = yTop + dy;

	    tBot = tTop + dt;
	    s    = 0.0;

	    glBegin(GL_TRIANGLE_STRIP);

	    for (side = 0; side < numSides; side++) {
		pt[0] = ringCoords[side][0];
		pt[2] = ringCoords[side][1];

		// Deal with normal
		norm.setValue(pt[0], 0.0, pt[2]);
		glNormal3fv(norm.getValue());

		// Point at bottom of section
		pt[1] = yBot;
		glTexCoord2f(s, tBot);
		glVertex3fv(pt.getValue());

		// Point at top of section
		pt[1] = yTop;
		glTexCoord2f(s, tTop);
		glVertex3fv(pt.getValue());
		s += ds;
	    }

	    // Join end of strip back to beginning
	    side = 0;
	    s = 1.0;
	    pt[0] = ringCoords[side][0];
	    pt[2] = ringCoords[side][1];

	    // Deal with normal
	    norm.setValue(pt[0], 0.0, pt[2]);
	    glNormal3fv(norm.getValue());

	    // Point at bottom of section
	    pt[1] = yBot;
	    glTexCoord2f(s, tBot);
	    glVertex3fv(pt.getValue());

	    // Point at top of section
	    pt[1] = yTop;
	    glTexCoord2f(s, tTop);
	    glVertex3fv(pt.getValue());
	    s += ds;

	    glEnd();

	    // Prepare for next section down
	    yTop = yBot;
	    tTop = tBot;
	}
    }

    // Draw top face as a series of concentric rings. The number of
    // rings is the same as the number of sections of the sides of the
    // cylinder.
    if (HAS_PART(curParts, TOP)) {
	norm.setValue(0.0, 1.0, 0.0);
	pt[1] = 1.0;

	glNormal3fv(norm.getValue());

	// Start at the outside and work in
	outerRadius = 1.0;
	dRadius     = -1.0 / numSections;
	for (section = numSections - 1; section >= 0; --section) {

	    innerRadius = outerRadius + dRadius;

	    // Innermost ring is treated as a triangle fan. This not
	    // only gets better shading (because the center vertex is
	    // sent), but also avoids the problem of having a polygon
	    // with too many vertices.
	    if (section == 0) {
		glBegin(GL_TRIANGLE_FAN);

		// Center point comes first
		pt[0] = pt[2] = 0.0;
		glTexCoord2f(0.5, 0.5);
		glVertex3fv(pt.getValue());

		// Send all vertices around ring. Go in reverse order
		// so that vertex ordering is correct
		for (side = numSides - 1; side >= 0; side--) {
		    pt[0] = outerRadius * ringCoords[side][0];
		    pt[2] = outerRadius * ringCoords[side][1];
		    glTexCoord2f(TOP_TEX_S(pt[0]), TOP_TEX_T(pt[2]));
		    glVertex3fv(pt.getValue());
		}
		// Send first vertex again
		pt[0] = outerRadius * ringCoords[numSides - 1][0];
		pt[2] = outerRadius * ringCoords[numSides - 1][1];
		glTexCoord2f(TOP_TEX_S(pt[0]), TOP_TEX_T(pt[2]));
		glVertex3fv(pt.getValue());

		glEnd();
	    }

	    // Other rings are triangle strips
	    else {
		glBegin(GL_TRIANGLE_STRIP);

		for (side = 0; side < numSides; side++) {
		    // Send points on outer and inner rings
		    pt[0] = outerRadius * ringCoords[side][0];
		    pt[2] = outerRadius * ringCoords[side][1];
		    glTexCoord2f(TOP_TEX_S(pt[0]), TOP_TEX_T(pt[2]));
		    glVertex3fv(pt.getValue());
		    pt[0] = innerRadius * ringCoords[side][0];
		    pt[2] = innerRadius * ringCoords[side][1];
		    glTexCoord2f(TOP_TEX_S(pt[0]), TOP_TEX_T(pt[2]));
		    glVertex3fv(pt.getValue());
		}

		// Join end of strip back to beginning
		pt[0] = outerRadius * ringCoords[0][0];
		pt[2] = outerRadius * ringCoords[0][1];
		glTexCoord2f(TOP_TEX_S(pt[0]), TOP_TEX_T(pt[2]));
		glVertex3fv(pt.getValue());
		pt[0] = innerRadius * ringCoords[0][0];
		pt[2] = innerRadius * ringCoords[0][1];
		glTexCoord2f(TOP_TEX_S(pt[0]), TOP_TEX_T(pt[2]));
		glVertex3fv(pt.getValue());

		glEnd();

		// Prepare for next ring
		outerRadius = innerRadius;
	    }
	}
    }

    // Draw bottom face the same way as the top
    if (HAS_PART(curParts, BOTTOM)) {
	norm.setValue(0.0, -1.0, 0.0);
	pt[1] = -1.0;

	glNormal3fv(norm.getValue());

	// Start at the outside and work in
	outerRadius = 1.0;
	dRadius     = -1.0 / numSections;
	for (section = numSections - 1; section >= 0; --section) {

	    innerRadius = outerRadius + dRadius;

	    // Innermost ring is drawn as a triangle fan. This not
	    // only gets better shading (because the center vertex is
	    // sent), but also avoids the problem of having a polygon
	    // with too many vertices.
	    if (section == 0) {
		glBegin(GL_TRIANGLE_FAN);

		// Center point comes first
		pt[0] = pt[2] = 0.0;
		glTexCoord2f(0.5, 0.5);
		glVertex3fv(pt.getValue());

		// Send all vertices around ring
		for (side = 0; side < numSides; side++) {
		    pt[0] = outerRadius * ringCoords[side][0];
		    pt[2] = outerRadius * ringCoords[side][1];
		    glTexCoord2f(BOT_TEX_S(pt[0]), BOT_TEX_T(pt[2]));
		    glVertex3fv(pt.getValue());
		}
		// Send first vertex again
		pt[0] = outerRadius * ringCoords[0][0];
		pt[2] = outerRadius * ringCoords[0][1];
		glTexCoord2f(BOT_TEX_S(pt[0]), BOT_TEX_T(pt[2]));
		glVertex3fv(pt.getValue());

		glEnd();
	    }

	    // Other rings are triangle strips
	    else {
		glBegin(GL_TRIANGLE_STRIP);

		// Go in reverse order so that vertex ordering is correct
		for (side = numSides - 1; side >= 0; side--) {
		    // Send points on outer and inner rings
		    pt[0] = outerRadius * ringCoords[side][0];
		    pt[2] = outerRadius * ringCoords[side][1];
		    glTexCoord2f(BOT_TEX_S(pt[0]), BOT_TEX_T(pt[2]));
		    glVertex3fv(pt.getValue());
		    pt[0] = innerRadius * ringCoords[side][0];
		    pt[2] = innerRadius * ringCoords[side][1];
		    glTexCoord2f(BOT_TEX_S(pt[0]), BOT_TEX_T(pt[2]));
		    glVertex3fv(pt.getValue());
		}

		// Join end of strip back to beginning
		side = numSides - 1;
		pt[0] = outerRadius * ringCoords[side][0];
		pt[2] = outerRadius * ringCoords[side][1];
		glTexCoord2f(BOT_TEX_S(pt[0]), BOT_TEX_T(pt[2]));
		glVertex3fv(pt.getValue());
		pt[0] = innerRadius * ringCoords[side][0];
		pt[2] = innerRadius * ringCoords[side][1];
		glTexCoord2f(BOT_TEX_S(pt[0]), BOT_TEX_T(pt[2]));
		glVertex3fv(pt.getValue());

		glEnd();

		// Prepare for next ring
		outerRadius = innerRadius;
	    }
	}
    }
}
