/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */

/************************************************************************
 *
 * File: %M%
 *
 * Copyright (c) 1994 by Silicon Graphics Inc., Mt. View, CA 94039
 *
 * Created:       Tue Jul 12 13:02:20 1994 by George Kyriazis
 * Last Modified: Tue Jul 12 13:40:54 1994 by George Kyriazis
 *
 * how to render a face set
 *
 ************************************************************************/

#ident "$Revision: 1.1 $"

#define private protected
#include "GlFaceSet.h"
#undef private

#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/elements/SoShapeHintsElement.h>
#include <Inventor/elements/SoGLCoordinateElement.h>
#include <Inventor/elements/SoComplexityElement.h>
#include <Inventor/bundles/SoTextureCoordinateBundle.h>
#include <Inventor/bundles/SoNormalBundle.h>

SO_NODE_SOURCE(GlFaceSet);


GlFaceSet::GlFaceSet()
{
    SO_NODE_CONSTRUCTOR(GlFaceSet);
    isBuiltIn = TRUE;
}

GlFaceSet::~GlFaceSet()
{
}


void
GlFaceSet::initClass()
{
    classTypeId = SoType::overrideType(SoFaceSet::getClassTypeId(),
				       &GlFaceSet::createInstance);
    parentFieldData = SoFaceSet::getFieldDataPtr();
}


void
GlFaceSet::GLRender(SoGLRenderAction *action)
{
    // See if we can bail-out early:
    if (! shouldGLRender(action))
	return;

    SoState *state = action->getState();

    // If faces might be concave, we have to send them through GLU:
    SoShapeHintsElement::VertexOrdering vo;
    SoShapeHintsElement::ShapeType st;
    SoShapeHintsElement::FaceType ft;
    SoShapeHintsElement::get(state, vo, st, ft);

    if (ft != SoShapeHintsElement::CONVEX) {
	// Use generatePrimitives for now...
	SoShape::GLRender(action);
	return;
    }

    SoNormalBundle		nb(action, TRUE);
    SoTextureCoordinateBundle	tcb(action, TRUE);
    const SoGLCoordinateElement	*ce;
    int				curVert, vert;
    int				face, numFaces;
    int				vertsInFace, vertsInPrevFace;
    Binding			normalBinding;

    ce = (const SoGLCoordinateElement *)
	SoCoordinateElement::getInstance(state);

    normalBinding   = getNormalBinding(action, &nb);

    curVert = (int) startIndex.getValue();

    if (normalBinding == OVERALL)
	nb.send(0);

    numFaces = numVertices.getNum();

    // Get the complexity element and decide how faces will be skipped
    // during rendering.
    float cmplxValue = SoComplexityElement::get(state);
    float delta = 1.8 * (0.5 - ((cmplxValue < 0.5) ? cmplxValue : 0.5));

    // For each face
    float fraction  = 0.0;
    vertsInPrevFace = -1;	// See comment below
    for (face = 0; face < numFaces; face++, fraction += delta) {

	// Figure out number of vertices in this face
	vertsInFace = (int) numVertices[face];
	if (vertsInFace == SO_FACE_SET_USE_REST_OF_VERTICES)
	    vertsInFace = (int) ce->getNum() - curVert;

	// Check to see if this face should be skipped due to complexity
	if (fraction >= 1.0) {
	    fraction -= 1.0;
	    curVert += vertsInFace;
	    continue;
	}

	// Send PER_FACE info:
	if (normalBinding == PER_FACE)
	    nb.send(face);

	// It's a little expensive to call glBegin() and glEnd()
	// repeatedly, so we avoid doing this when possible. It's
	// likely that faces have 3 or 4 vertices, and it is also
	// likely that a contiguous bunch of faces have the same
	// number of vertices. We need to do begin/end only if the
	// number of vertices changes or there are more than 4
	// vertices in a face (so we have to use GL_POLYGON, which
	// handles only 1 polygon at a time).
	if (vertsInFace != vertsInPrevFace) {
	    if (vertsInPrevFace >= 0)
		glEnd();
	    glBegin(vertsInFace == 3 ? GL_TRIANGLES :
		    vertsInFace == 4 ? GL_QUADS     : GL_POLYGON);

	    // Set vertsInPrevFace for next time. If there's more than
	    // 4 vertices in the current face, set it to 0 so that (1)
	    // the next face will have a different number of vertices
	    // (presumably), and (2) the glEnd() will be called above.
	    vertsInPrevFace = (vertsInFace > 4 ? 0 : vertsInFace);
	}

	for (vert = 0; vert < vertsInFace; vert++) {

	    // Send PER_VERTEX info:
	    if (normalBinding == PER_VERTEX)
		nb.send(curVert);

	    // And texture coordinates, if necessary:
	    if (tcb.needCoordinates())
		tcb.send(curVert);

	    // Finally, send the vertex:
	    ce->send(curVert);

	    curVert++;
	}
    }

    glEnd();
}
