/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */

/************************************************************************
 *
 * File: %M%
 *
 * Copyright (c) 1994 by Silicon Graphics Inc., Mt. View, CA 94039
 *
 * Created:       Tue Jul 12 13:15:50 1994 by George Kyriazis
 * Last Modified: Tue Jul 12 13:41:03 1994 by George Kyriazis
 *
 * how to render an indexed face set
 *
 ************************************************************************/

#ident "$Revision: 1.1 $"

#define private protected
#include "GlIndexedFaceSet.h"
#undef private

#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/elements/SoShapeHintsElement.h>
#include <Inventor/elements/SoGLCoordinateElement.h>
#include <Inventor/elements/SoComplexityElement.h>
#include <Inventor/bundles/SoTextureCoordinateBundle.h>
#include <Inventor/bundles/SoNormalBundle.h>


SO_NODE_SOURCE(GlIndexedFaceSet);


GlIndexedFaceSet::GlIndexedFaceSet()
{
    SO_NODE_CONSTRUCTOR(GlIndexedFaceSet);
    isBuiltIn = TRUE;
}

GlIndexedFaceSet::~GlIndexedFaceSet()
{
}


void
GlIndexedFaceSet::initClass()
{
    classTypeId = SoType::overrideType(SoIndexedFaceSet::getClassTypeId(),
				       &GlIndexedFaceSet::createInstance);
    parentFieldData = SoIndexedFaceSet::getFieldDataPtr();
}


void
GlIndexedFaceSet::GLRender(SoGLRenderAction *action)
{
    // See if we can bail-out early:
    if (! shouldGLRender(action))
	return;

    SoState *state = action->getState();

    // If faces might be concave, we have to send them through GLU:
    SoShapeHintsElement::VertexOrdering vo;
    SoShapeHintsElement::ShapeType st;
    SoShapeHintsElement::FaceType ft;
    SoShapeHintsElement::get(state, vo, st, ft);

    if (ft != SoShapeHintsElement::CONVEX) {
	// Use generatePrimitives for now...
	SoShape::GLRender(action);
	return;
    }

    // Rendering code:  a lot like generatePrimitives...
    SoNormalBundle		nb(action, TRUE);
    SoTextureCoordinateBundle	tcb(action, TRUE);
    const SoGLCoordinateElement	*ce;
    int				curVert, curIndex, curFace;
    int				numIndices, vertsInFace, vertsInPrevFace;
    const long			*coordIndices;
    const long			*normIndices, *texCoordIndices;
    Binding			normalBinding;
    SbBool			texCoordsIndexed;

    ce = (const SoGLCoordinateElement *)
	SoCoordinateElement::getInstance(state);

    texCoordsIndexed = areTexCoordsIndexed(action);
    normalBinding    = getNormalBinding(action, &nb);

    curFace = curVert = curIndex = 0;

    numIndices      = coordIndex.getNum();
    coordIndices    = coordIndex.getValues(0);
    normIndices     = normalIndex.getValues(0);
    texCoordIndices = textureCoordIndex.getValues(0);

    // Check for special case of 1 index of SO_END_FACE_INDEX. This
    // means that coord indices are to be used for materials, normals,
    // or texture coords as well
    if (normalIndex.getNum() == 1 && normIndices[0] == SO_END_FACE_INDEX) 
	normIndices = coordIndices;
    if (textureCoordIndex.getNum() == 1 &&
	texCoordIndices[0] == SO_END_FACE_INDEX)
	texCoordIndices = coordIndices;

    if (normalBinding == OVERALL)
	nb.send(0);

    // Get the complexity element and decide how faces will be skipped
    // during rendering.
    float cmplxValue = SoComplexityElement::get(state);
    float delta = 1.8 * (0.5 - ((cmplxValue < 0.5) ? cmplxValue : 0.5));

    // Step through all the coordinate indices, building faces out
    // of them, until we run out of coordinate indices.
    float fraction = 0.0;
    vertsInPrevFace = -1;	// See comment below
    while (curIndex < numIndices) {

	vertsInFace = getNumVerts(curIndex);

	// Forget about faces with less than three vertices...
	if (vertsInFace < 3) {
	    curIndex += vertsInFace+1;
	    curVert += vertsInFace;
	    curFace++;
	    continue;
	}

	// Check to see whether to skip this face due to low complexity
	fraction += delta;
	if (fraction >= 1.0) {
	    fraction -= 1.0;
	    curIndex += vertsInFace+1;
	    curVert += vertsInFace;
	    curFace++;
	    continue;
	}

	// Send PER_FACE info:
	if (normalBinding == PER_FACE)
	    nb.send(curFace);

	// And PER_FACE_INDEXED:
	if (normalBinding == PER_FACE_INDEXED)
	    nb.send((int) normIndices[curFace]);

	// It's a little expensive to call glBegin() and glEnd()
	// repeatedly, so we avoid doing this when possible. It's
	// likely that faces have 3 or 4 vertices, and it is also
	// likely that a contiguous bunch of faces have the same
	// number of vertices. We need to do begin/end only if the
	// number of vertices changes or there are more than 4
	// vertices in a face (so we have to use GL_POLYGON, which
	// handles only 1 polygon at a time).

	// This tests to see how many vertices are in the next face.
	// It makes a couple of assumptions in order to be as fast as
	// possible.
	vertsInFace = (coordIndices[curIndex+3] == SO_END_FACE_INDEX ? 3 :
		       coordIndices[curIndex+4] == SO_END_FACE_INDEX ? 4 : 5);

	if (vertsInFace != vertsInPrevFace) {
	    if (vertsInPrevFace >= 0)
		glEnd();
	    glBegin(vertsInFace == 3 ? GL_TRIANGLES :
		    vertsInFace == 4 ? GL_QUADS     : GL_POLYGON);

	    // Set vertsInPrevFace for next time. If there's more than
	    // 4 vertices in the current face, set it to 0 so that (1)
	    // the next face will have a different number of vertices
	    // (presumably), and (2) the glEnd() will be called above.
	    vertsInPrevFace = (vertsInFace > 4 ? 0 : vertsInFace);
	}

	// Loop through all vertices of current face
	while (curIndex < numIndices &&
	       coordIndices[curIndex] != SO_END_FACE_INDEX) {

	    // Send PER_VERTEX info:
	    if (normalBinding == PER_VERTEX)
		nb.send(curVert);

	    // And PER_VERTEX_INDEXED:
	    if (normalBinding == PER_VERTEX_INDEXED)
		nb.send((int) normIndices[curIndex]);

	    // And texture coordinates, if necessary:
	    if (tcb.needCoordinates()) {
		if (texCoordsIndexed)
		    tcb.send((int) texCoordIndices[curIndex]);
		else
		    tcb.send(curVert);
	    }
	    // Finally, send the vertex:
	    ce->send((int) coordIndices[curIndex]);

	    //
	    // Increment per-vertex stuff
	    //
	    curVert++;
	    curIndex++;
	}

	//
	// Increment per-face stuff
	//
	curIndex++; 	// Skip over the END_FACE_INDEX
	curFace++;
    }

    if (vertsInPrevFace >= 0)
	glEnd();
}
