/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */

/************************************************************************
 *
 * File: %M%
 *
 * Copyright (c) 1994 by Silicon Graphics Inc., Mt. View, CA 94039
 *
 * Created:       Tue Jul 12 11:58:39 1994 by George Kyriazis
 * Last Modified: Tue Jul 12 13:40:32 1994 by George Kyriazis
 *
 * how to render a sphere
 *
 ************************************************************************/

#ident "$Revision: 1.1 $"

#define private protected
#include "GlSphere.h"
#undef private

#include <Inventor/actions/SoGLRenderAction.h>


SO_NODE_SOURCE(GlSphere);


// Computes S and T texture coordinates from point on surface
#define COMPUTE_S_T(point, s, t)					      \
    s = atan2f(point[0], point[2]) * .159 + .5;				      \
    t = atan2f(point[1],						      \
	       sqrtf(point[0]*point[0] + point[2]*point[2])) * .318 + .5

;
// Adjusts S texture coordinate in unstable areas
#define ADJUST_S(s, octant)						      \
    if (s < .001 && (octant == 1 || octant == 3))			      \
	s = 1.0;							      \
    else if (s > .999 && (octant == 5 || octant == 7))			      \
	s = 0.0


GlSphere::GlSphere()
{
    SO_NODE_CONSTRUCTOR(GlSphere);
    isBuiltIn = TRUE;
}

GlSphere::~GlSphere()
{
}


void
GlSphere::initClass()
{
    classTypeId = SoType::overrideType(SoSphere::getClassTypeId(),
				       &GlSphere::createInstance);
    parentFieldData = SoSphere::getFieldDataPtr();
}


void
GlSphere::GLRender(SoGLRenderAction *action)
{
    float	rad;

    // First see if the object is visible and should be rendered now
    if (! shouldGLRender(action))
	return;

    // Scale world if radius is not 1
    rad = (radius.isIgnored() ? 1.0 : radius.getValue());

    if (rad != 1.0) {
	glPushMatrix();
	glScalef(rad, rad, rad);
    }

    // Render the sphere. 
    GLRenderGeneric(action);

    if (rad != 1.0)
	glPopMatrix();
}


void
GlSphere::GLRenderGeneric( SoGLRenderAction *action )
{
    int		i, j, k, s_x, s_y, s_z, order, octant;
    float	botWidth, topWidth, yTop, yBot, tmp;
    SbVec3f	vec;
    int		depth;
    float	s, t, sAvg;

    // Compute depth based on complexity
    depth = computeDepth(action);


    for (octant = 0; octant < 8; octant++) {
	s_x = -(((octant & 01) << 1) - 1);
	s_y = -( (octant & 02)       - 1);
	s_z = -(((octant & 04) >> 1) - 1);
	order = s_x * s_y * s_z;

	for (i = 0; i < depth - 1; i++) {
	    yBot = (float) i      / depth;
	    yTop = (float)(i + 1) / depth;

	    botWidth = 1.0 - yBot;
	    topWidth = 1.0 - yTop;

	    glBegin(GL_TRIANGLE_STRIP);

	    for (j = 0; j < depth - i; j++) {

		// First vertex
		k = order > 0 ? depth - i - j : j;
		tmp = (botWidth * k) / (depth - i);
		vec.setValue(s_x * tmp, s_y * yBot, s_z * (botWidth - tmp));
		vec.normalize();

		COMPUTE_S_T(vec, s, t);
		glTexCoord2f(s, t);
		glNormal3fv(vec.getValue());
		glVertex3fv(vec.getValue());

		// Second vertex
		k = order > 0 ? (depth - i - 1) - j : j;
		tmp = (topWidth * k) / (depth - i - 1);
		vec.setValue(s_x * tmp, s_y * yTop, s_z * (topWidth - tmp));
		vec.normalize();

		COMPUTE_S_T(vec, s, t);
		ADJUST_S(s, octant);
		glTexCoord2f(s, t);
		glNormal3fv(vec.getValue());
		glVertex3fv(vec.getValue());
	    }

	    // Last vertex
	    k = order > 0 ? depth - i - j : j;
	    tmp = (botWidth * k) / (depth - i);
	    vec.setValue(s_x * tmp, s_y * yBot, s_z * (botWidth - tmp));
	    vec.normalize();

	    COMPUTE_S_T(vec, s, t);
	    ADJUST_S(s, octant);
	    glTexCoord2f(s, t);
	    glNormal3fv(vec.getValue());
	    glVertex3fv(vec.getValue());

	    glEnd();
	}

	// Handle the top/bottom polygons specially, to avoid divide by zero
	glBegin(GL_TRIANGLE_STRIP);

	yBot = (float) i / depth;
	yTop = 1.0;
	botWidth = 1 - yBot;

	// First cap vertex
	if (order > 0)
	    vec.setValue(0.0, s_y * yBot, s_z * botWidth);
	else
	    vec.setValue(s_x * botWidth, s_y * yBot, 0.0);
	vec.normalize();

	COMPUTE_S_T(vec, s, t);
	ADJUST_S(s, octant);
	sAvg = s;
	glTexCoord2f(s, t);
	glNormal3fv(vec.getValue());
	glVertex3fv(vec.getValue());

	// Second cap vertex
	if (order > 0)
	    vec.setValue(s_x * botWidth, s_y * yBot, 0.0);
	else
	    vec.setValue(0.0, s_y * yBot, s_z * botWidth);
	vec.normalize();

	COMPUTE_S_T(vec, s, t);
	ADJUST_S(s, octant);
	sAvg = (sAvg + s) / 2;
	glTexCoord2f(s, t);
	glNormal3fv(vec.getValue());
	glVertex3fv(vec.getValue());

	// Third cap vertex
	vec.setValue(0.0, s_y, 0.0);

	s = sAvg;
	t = s_y * .5 + .5;
	glTexCoord2f(s, t);
	glNormal3fv(vec.getValue());
	glVertex3fv(vec.getValue());

	glEnd();
    }
}
