/*
 * Copyright (C) 1990, 1991, 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
/*
 * Copyright (C) 1990,91   Silicon Graphics, Inc.
 *
 _______________________________________________________________________
 ______________  S I L I C O N   G R A P H I C S   I N C .  ____________
 |
 |   $Revision: 1.1 $
 |
 |   Classes:
 |	SbDictEntry, SbDict
 |
 |   Author(s)		: Nick Thompson, Paul S. Strauss, Dave Immel
 |
 |   Notes:
 |	Dictionaries are maintained as external hash tables of
 |	SbDictEntry instances.
 |
 ______________  S I L I C O N   G R A P H I C S   I N C .  ____________
 _______________________________________________________________________
 */

#include <Inventor/SbDict.h>

////////////////////////////////////////////
//
// This structure is used to implement makePList()
//

struct SbDictListThing {
    SbPList	*keyList;
    SbPList	*valueList;
};

//
////////////////////////////////////////////

//
// Constructor - empties all buckets in the hash table
//

SbDict::SbDict( int entries )
{
    tableSize=entries;
    buckets=new SbDictEntry *[tableSize];
    for (int i = 0; i < tableSize; i++)
	buckets[i] = NULL;
}

//
// Destructor - clears all buckets, frees up bucket table.
//

SbDict::~SbDict()
{
    clear();
    delete [] buckets;
}

//
// Adds one entry to the end of the PLists accessed through "data"
//

void
SbDict::addEntryToPLists(u_long key, void *value, void *data)
{
    SbDictListThing	*lists = (SbDictListThing *) data;

    lists->keyList->append((void *) key);
    lists->valueList->append(value);
}

//
// Calls given routine (passing value) for each entry in dictionary.
// The order of entries is not guaranteed to mean anything.
//

void
SbDict::applyToAll(void (*rtn)(u_long key, void *value) )
{
    int		i;
    SbDictEntry	*entry;

    // Call rtn for each entry in dict
    for (i = 0; i < tableSize; i++) {

	for (entry = buckets[i]; entry != NULL; entry = entry->next)
	    (*rtn)(entry->key, entry->value);
    }
}

//
// Calls given routine (passing value,data) for each entry in dictionary.
// The order of entries is not guaranteed to mean anything.
//

void
SbDict::applyToAll(void (*rtn)(u_long key, void *value, void *data),
		   void *data )
{
    int		i;
    SbDictEntry	*entry;

    // Call rtn for each entry in dict
    for (i = 0; i < tableSize; i++) {

	for (entry = buckets[i]; entry != NULL; entry = entry->next)
	    (*rtn)(entry->key, entry->value, data);
    }
}

//
// Removes all entries from dictionary.
//

void
SbDict::clear()
{
    int		i;
    SbDictEntry	*entry, *nextEntry;

    // Free up each entry in dict
    for (i = 0; i < tableSize; i++) {

	for (entry = buckets[i]; entry != NULL; entry = nextEntry) {
	    nextEntry = entry->next;
	    delete entry;
	}
	buckets[i] = NULL;
    }
}

//
// Enters a key,value pair into the dictionary. Overwrites entry and
// returns FALSE if key already exists.
//

SbBool
SbDict::enter(u_long key, void *value)
{
    SbDictEntry		*&entry = findEntry(key);

    if (entry == NULL) {
	entry = new SbDictEntry(key, value);
	entry->next = NULL;
	return TRUE;
    }
    else {
	entry->value = value;	// Overwrites old value
	return FALSE;
    }
}

//
// Finds entry with given key, setting value to point to value.
// Returns FALSE if no such entry.
//

SbBool
SbDict::find(u_long key, void *&value) const
{
    SbDictEntry		*&entry = findEntry(key);

    if (entry == NULL) {
	value = NULL;
	return FALSE;
    }
    else {
	value = entry->value;
	return TRUE;
    }
}

//
// Returns reference to entry with given key, or NULL if there is none.
//

SbDictEntry *&
SbDict::findEntry(u_long key) const
{
    SbDictEntry		**entry;

    entry = &buckets[key % tableSize];

    while (*entry != NULL) {
	if ((*entry)->key == key)
	    break;
	entry = &(*entry)->next;
    }
    return *entry;
}

//
// Makes two SbPLists, one for keys and the other for values.
// The order of entries is not guaranteed to mean anything.
//

void
SbDict::makePList(SbPList &keys, SbPList &values)
{
    SbDictListThing	lists;

    lists.keyList = &keys;
    lists.valueList = &values;

    keys.truncate(0);
    values.truncate(0);

    applyToAll(SbDict::addEntryToPLists, (void *) &lists);
}

//
// Removes the entry with the given key. Returns FALSE if no such entry.
//

SbBool
SbDict::remove(u_long key)
{
    SbDictEntry		*&entry = findEntry(key);
    SbDictEntry		*tmp;

    if (entry == NULL)
	return FALSE;
    else {
	tmp = entry;
	entry = entry->next;
	delete tmp;
	return TRUE;
    }
}
