/*
 * Copyright (C) 1990, 1991, 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
/*
 * Copyright (C) 1990,91   Silicon Graphics, Inc.
 *
 _______________________________________________________________________
 ______________  S I L I C O N   G R A P H I C S   I N C .  ____________
 |
 |   $Revision: 1.1 $
 |
 |   Classes:
 |	SbNameEntry, SbName
 |
 |   Author(s)		: Nick Thompson, Paul S. Strauss
 |
 |   Notes:
 |	SbNames are stored in a hash table of SbNameEntries.
 |
 ______________  S I L I C O N   G R A P H I C S   I N C .  ____________
 _______________________________________________________________________
 */

#include <Inventor/SbString.h>

//////////////////////////////////////////////////////////////////////////////
//
// SbNameEntry class. Instances of this class are stored in a hash
// table that is used to find names quickly. All of these methods are
// private. Some are used by SbName, which is a friend class.
//
//////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////
//
// This structure contains a chunk of memory in which strings are
// stored. Doing this reduces memory fragmentation.
//
////////////////////////////////////////////

// The size of a chunk is about a page (4K) minus room for other stuff
// in the structure
#define CHUNK_SIZE	4000

struct SbNameChunk {
    char		mem[CHUNK_SIZE];	// Chunk o'memory
    char		*curByte;		// Current byte of chunk
    int			bytesLeft;		// Bytes left in chunk
    struct SbNameChunk	*next;			// Pointer to next chunk
};

int		SbNameEntry::nameTableSize; // Number of buckets in name table
SbNameEntry **	SbNameEntry::nameTable;	    // Array of name entries
struct SbNameChunk *SbNameEntry::chunk;	    // Chunk of memory for string store

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Initializes SbNameEntry class. This is done only once.
//
// Use: private

void
SbNameEntry::initClass()
//
////////////////////////////////////////////////////////////////////////
{
    int i;

    // Allocate and empty name hash table

    nameTableSize = 1999;	// About 2000 (primed)
    nameTable	  = new SbNameEntry *[nameTableSize];

    for (i = 0; i < nameTableSize; i++)
	nameTable[i] = NULL;

    chunk = NULL;
}

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Inserts a string into the hash table.
//
// Use: private

const SbNameEntry *
SbNameEntry::insert(const char *s)
//
////////////////////////////////////////////////////////////////////////
{
    u_long		h = SbString::hash(s);
    u_long		i;
    SbNameEntry		*entry;
    SbNameEntry		*head;

    if (nameTableSize == 0)
	initClass();

    i = h % nameTableSize;
    entry = head = nameTable[i];

    // Look for entry with same string
    while (entry != NULL) {
	if (entry->hashValue == h && entry->isEqual(s))
	    break;
	entry = entry->next;
    }

    // If not there, create one
    if (entry == NULL) {

	int len = strlen(s) + 1;

	// If there's no way to fit this string in a chunk
	if (len >= CHUNK_SIZE)
	    s = strdup(s);

	else {

	    // Make sure there's room in the current chunk; create a
	    // new chunk if necessary.
	    if (chunk == NULL || chunk->bytesLeft < len) {
		struct SbNameChunk	*newChunk = new SbNameChunk;

		newChunk->curByte   = newChunk->mem;
		newChunk->bytesLeft = CHUNK_SIZE;
		newChunk->next      = chunk;

		chunk = newChunk;
	    }

	    // Store string in chunk
	    strcpy(chunk->curByte, s);
	    s = chunk->curByte;

	    chunk->curByte   += len;
	    chunk->bytesLeft -= len;
	}

	entry = new SbNameEntry(s, h, head);
	nameTable[i] = entry;
    }

    return entry;
}

//////////////////////////////////////////////////////////////////////////////
//
// SbName class.
//
//////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Default constructor
//
// Use: private

SbName::SbName()
//
////////////////////////////////////////////////////////////////////////
{
    entry = SbNameEntry::insert("");
}
